# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas.motl@ip-tools.org>
import os
import sys
import json
import timeit
import logging
from appdirs import user_data_dir


def to_list(obj):
    """Convert an object to a list if it is not already one"""
    # stolen from cornice.util
    if not isinstance(obj, (list, tuple)):
        obj = [obj, ]
    return obj


def read_list(data, separator=u','):
    if data is None:
        return []
    result = list(map(lambda x: x.strip(), data.split(separator)))
    if len(result) == 1 and not result[0]:
        result = []
    return result


def boot_logging(options=None):
    log_level = logging.INFO
    if options and options.get('debug'):
        log_level = logging.DEBUG
    setup_logging(level=log_level)


def setup_logging(level=logging.INFO):
    log_format = '%(asctime)-15s [%(name)-20s] %(levelname)-7s: %(message)s'
    logging.basicConfig(
        format=log_format,
        stream=sys.stderr,
        level=level)


def normalize_options(options):
    normalized = {}
    for key, value in options.items():
        key = key.strip('--<>')
        normalized[key] = value
    return normalized


def read_numbersfile(filename):
    numbers = open(filename, 'r').readlines()
    numbers = map(str.strip, numbers)
    numbers = filter(lambda number: not number.startswith('#'), numbers)
    return numbers


class SmartException(Exception):

    def __init__(self, message, **kwargs):

        # Call the base class constructor with the parameters it needs
        super(SmartException, self).__init__(message)

        # Stuff more things into the exception object
        self.more_info = kwargs


class PersistentConfiguration(dict):

    def __init__(self, appname=None, configfile=None):
        self.store = {}
        if configfile:
            self.configfile = configfile
        else:
            self.configpath = user_data_dir(appname=appname, appauthor=False)
            if not os.path.exists(self.configpath):
                os.makedirs(self.configpath)
            self.configfile = os.path.join(self.configpath, 'config.json')
        self.setup()

    def setup(self):
        # json_store is not compatible with Python 3
        #self.store = json_store.open(self.configfile)
        if os.path.exists(self.configfile):
            with open(self.configfile, 'r') as f:
                self.store = json.load(f)

    def has_key(self, key):
        return self.store.has_key(key)

    def __getitem__(self, key):
        return self.store[key]

    def __setitem__(self, key, value):
        self.store[key] = value
        self.sync()

    def sync(self):
        # json_store is not compatible with Python 3
        #self.store.sync()
        with open(self.configfile, 'w') as f:
            json.dump(self.store, f, indent=4)


class StopWatch:

    def __init__(self):
        self.start()

    def start(self):
        self.starttime = self.now()

    def now(self):
        return timeit.default_timer()

    def elapsed(self):
        duration = round(self.now() - self.starttime, 2)
        return duration



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas.motl@ip-tools.org>

class PatBaseException(Exception):
    pass

class PatBaseRequestError(PatBaseException):
    pass

class PatBaseLoginError(PatBaseException):
    pass

class PatBaseQueryError(PatBaseException):
    pass

class PatBaseResultError(PatBaseException):
    pass



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas.motl@ip-tools.org>
import json
import logging
from docopt import docopt
from patbase import __version__
from patbase.client import PatBaseClient
from patbase.util import boot_logging, normalize_options, PersistentConfiguration

"""
Python command line client for accessing the Minesoft PatBase REST API.

See also:
- https://www.patbase.com/rest/
- https://www.patbase.com/rest/PatBaseRestAPI.pdf
"""

logger = logging.getLogger(__name__)

APP_NAME = 'patbase'

def run():
    """
    Usage:
        {program} login --username=<username> --password=<password>
        {program} search <expression> [--username=<username> --password=<password>]
        {program} info
        {program} --version
        {program} (-h | --help)

    Examples:

        # Display published application by publication number in XML format
        {program} login --username=test@example.org --password=secret

        # Search for documents matching "Space Shuttle" in all fulltext fields and display JSON response
        {program} search 'FT=(Space Shuttle)'

    """

    # Use generic commandline options schema and amend with current program name
    commandline_schema = (run.__doc__).format(program=APP_NAME)

    # Read commandline options
    options = normalize_options(docopt(commandline_schema, version=APP_NAME + ' ' + __version__))

    # Start logging subsystem
    boot_logging(options)

    # The configuration file for storing API credentials
    config = PersistentConfiguration(appname='minesoft-patbase-client')

    if options['login']:

        # Create API client
        client = PatBaseClient(username=options['username'], password=options['password'])
        client.login()

        # Save credentials for subsequent operations
        logger.info('Saving login credentials to {}'.format(config.configfile))
        config['api_credentials'] = {
            'username': options['username'],
            'password': options['password'],
        }

    elif options['search']:

        # Optionally load credentials from configuration file
        username, password = options['username'], options['password']
        if not username or not password:
            api_credentials = config['api_credentials']
            username, password = api_credentials['username'], api_credentials['password']

        # Create API client
        client = PatBaseClient(username=username, password=password)
        query = client.query(options['expression'])
        result = client.searchresults(query, offset=0, limit=20)
        print(json.dumps(result.asdict()))



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()

requires = [
    'attrs>=17.3.0,<18.0.0',
    'docopt==0.6.2',
    'appdirs>=1.4.3,<2.0.0',
    'requests>=2.18.4',
    'jsonpointer>=1.12,<2.0.0',
    'maya==0.5.0',
]

test_requires = [
]

setup(name='minesoft-patbase-client',
      version='0.1.0',
      description='minesoft-patbase-client is a client library for accessing the Minesoft PatBase REST API',
      long_description=README,
      license="AGPL 3, EUPL 1.2",
      classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Legal Industry",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Archiving",
        "Topic :: Text Processing",
        "Topic :: Utilities",
      ],
      author='Andreas Motl',
      author_email='andreas.motl@ip-tools.org',
      url='https://github.com/ip-tools/minesoft-patbase-client',
      keywords='patent information minesoft patbase rest api client bulk data download import interface '
               'research search opendata xml json',
      packages=find_packages(),
      include_package_data=True,
      package_data={
      },
      zip_safe=False,
      test_suite='nose.collector',
      install_requires=requires,
      tests_require=test_requires,
      extras_require={
          'release': [
              'bumpversion==0.5.3',
              'twine==1.9.1',
              'keyring==10.4.0',
          ],
          'documentation': [
              'Sphinx==1.6.4',
              'sphinx_rtd_theme==0.2.5b1',
          ],
      },
      dependency_links=[
      ],

      entry_points={
        'console_scripts': [
            'patbase = patbase.command:run',
        ],
      },

    )



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas.motl@ip-tools.org>
#
# HTTP client for accessing the Minesoft PatBase REST API.
# See also: https://www.patbase.com/rest/
#
import attr
import json
import maya
import logging
import requests

from patbase.exceptions import PatBaseRequestError, PatBaseLoginError, PatBaseQueryError, PatBaseResultError
from patbase.util import StopWatch

logger = logging.getLogger(__name__)


@attr.s
class PatBaseClientBase(object):

    def logout(self):
        # Signal unauthenticated state
        self.authenticated = False

    def ensure_login(self):
        if not self.authenticated:
            self.login()

    def process_response(self, response, on_empty='raise'):

        # Read response content
        payload = response.content

        # The server seems to send an UTF-8 byte order mark, compensate for that.
        # https://en.wikipedia.org/wiki/Byte_order_mark#UTF-8
        try:
            payload = payload.decode('utf-8-sig')
        except:
            pass

        # Sanity checks
        if payload == '' and on_empty == 'raise':
            raise PatBaseRequestError('Empty response from PatBase API')

        # Decode JSON payload
        data = json.loads(payload)

        return data


@attr.s
class PatBaseClient(PatBaseClientBase):

    # The API credentials for authenticating with the service
    username = attr.ib()
    password = attr.ib()

    # The URL to the HTTP API endpoint
    url = attr.ib(default='https://www.patbase.com/rest/api.php')

    # The maximum number of results returned is 100,000
    query_max_results = attr.ib(default=100000)

    # Whether the client instance was able to authenticate properly
    authenticated = attr.ib(default=False)

    # Whether to enable/disable X.509 certificate verification when performing the TLS handshake
    tls_verify = attr.ib(default=True)

    # The HTTP client session instance from the "requests" module
    session = attr.ib(default=None)

    def login(self):

        # Signal unauthenticated state
        self.authenticated = False

        # Create HTTP session object which will keep track of cookies
        self.session = requests.Session()

        # Query parameters
        params = {'method': 'login', 'userid': self.username, 'password': self.password}

        # Send request
        response = self.session.get(self.url, params=params)

        # Read response
        data = self.process_response(response)

        # Check response
        if 'LOGIN_TO_API' in data and data['LOGIN_TO_API'] == 'OK':
            logger.info('PatBase API login succeeded')
        else:
            raise PatBaseLoginError('PatBase API login failed, please check your credentials. The error was: {}'.format(data))

        # Signal authenticated state
        self.authenticated = True

    def query(self, expression):

        # Start stopwatch
        stopwatch = StopWatch()

        # Make sure we are authenticated with the API
        self.ensure_login()

        # Report what we will be doing
        logger.info('PatBase API query using expression "{}"'.format(expression))

        # Query parameters
        params = {'method': 'query', 'query': expression}

        # Send request
        response = self.session.get(self.url, params=params)

        # Read response
        data = self.process_response(response)

        # Check response
        if 'Results' in data and 'QueryKey' in data:
            date = response.headers.get('Date')
            duration = stopwatch.elapsed()
            if date:
                date = maya.parse(date).rfc3339()
            outcome = PatBaseQuery(
                expression=expression,
                results=data['Results'],
                date=date,
                duration=duration,
                querykey=data['QueryKey'],
            )
            logger.info('PatBase API query succeeded with: {}'.format(outcome))
            return outcome
        else:
            raise PatBaseQueryError('PatBase API query failed, please check your expression. The error was: {}'.format(data))

    def searchresults(self, query, offset=None, limit=None, sort=None):

        # Make sure we are authenticated with the API
        self.ensure_login()

        # Report what we will be doing
        logger.info('PatBase API searchresults using query: {}'.format(query))

        # Query parameters
        params = {'method': 'searchresults', 'querykey': query.querykey}
        if offset is not None:
            params['from'] = offset
            if limit is not None:
                params['to'] = offset + limit

        # TODO: Add knowledge about mapping
        if sort is not None:
            params['sortorder'] = sort

        # Send request
        response = self.session.get(self.url, params=params)

        # Read response
        data = self.process_response(response)

        # Check response
        if 'Families' in data:
            outcome = PatBaseResult(query=query, records=data['Families'])
            logger.info('PatBase API searchresults succeeded with: {}'.format(outcome))
            return outcome
        else:
            raise PatBaseResultError(
                'PatBase API searchresults failed. The error was: {}'.format(data))


@attr.s
class PatBaseQuery(object):

    # Search expression string
    expression = attr.ib()

    # Total result count
    results = attr.ib()

    # Timestamp of query response
    date = attr.ib()

    # Duration of query operation
    duration = attr.ib()

    # QueryKey token for retrieving results
    querykey = attr.ib()


@attr.s
class PatBaseResult(object):

    # PatBaseQuery instance
    query = attr.ib()

    # List of result items
    records = attr.ib(default=attr.Factory(list), repr=False)

    def asdict(self):
        return attr.asdict(self)
